/* $Id: nwsclient.c,v 1.4 2004/03/05 22:01:34 jhayes Exp $ */
/*
 * Copyright  2004 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <oshseed.h>
#include <stdlib.h> /* atoi free malloc */
#include <string.h> /* memcpy memset strchr strlen */
#include <time.h>   /* time */
#define ASFMT_SHORT_NAMES
#include "formatseed.h"
#define ASIP_SHORT_NAMES
#include "ipseed.h"
#include "nwsclient.h"


#define HostToNetwork HomogenousConvertHostToNetwork
#define NetworkToHost HomogenousConvertNetworkToHost


typedef enum {
  /* Specific to memories */
  STORE_STATE = 100, STATE_SORTED,
  FETCH_STATE, STATE_FETCHED,
  AUTOFETCH_BEGIN, AUTOFETCH_ACK,
  MEMORY_CLEAN, MEMORY_CLEANED,
  MEMORY_LOGDEST, MEMORY_LOGDEST_ACK,
  MEMORY_FAILED = 199,
  /* Applicable to all hosts */
  HOST_TEST = 200, HOST_TEST_RESULT,
  HOST_DIAGNOSTICS, HOST_DIAGNOSTICS_ACK,
  HOST_DIE, HOST_DYING,
  HOST_REGISTER, HOST_REGISTERED,
  HOST_REFUSED = 298, HOST_FAILED,
  /* Specific to periodic control */
  PERIODIC_SERIES = 300,
  /* Specific to clique control */
  CLIQUE_ACTIVATE = 400, CLIQUE_DIE, CLIQUE_TOKEN_ACK, CLIQUE_TOKEN_FWD,
  CLIQUE_SERIES,
  /* Specific to forecasters */
  FORE_DATA = 500, FORE_FORECAST, FORE_SERIES,
  METHODS_ASK, METHODS_TELL,
  FORE_FAIL = 599,
  /* Specific to sensors */
  ACTIVITY_START = 600, ACTIVITY_STARTED,
  ACTIVITY_STOP, ACTIVITY_STOPPED,
  SENSOR_FAILED = 699,
  /* Specific to name servers */
  NS_REGISTER = 700, NS_REGISTERED,
  NS_SEARCH, NS_SEARCHED,
  NS_UNREGISTER, NS_UNREGISTERED,
  NS_FAILED = 799,
  /* Specific to skills */
  TCP_BW_REQ = 800, TCP_HANDSHAKE = 411,
  MESSAGE_COUNT
} NwsMessages;


static const int NWS_VERSION = 0x02000000;
#define STATE_NAME_SIZE 128
#define FORECAST_NAME_SIZE STATE_NAME_SIZE


/*
 * Receives a message on #connection#.  If #expected# is not received, returns
 * 0; otherwise, returns 1 and stores any accompanying message body in #body#
 * if #body# is not NULL.
 */
static int
ReceiveMessage(Socket connection,
               NwsMessages expected,
               char **body) {

  void *data = NULL;
  int dataSize;
  int message;
  char networkHeader[4 + 4 + 4];

  Receive(connection, networkHeader, sizeof(networkHeader));
  NetworkToHost(&message, networkHeader + 4, INT_TYPE, 1);
  NetworkToHost(&dataSize, networkHeader + 8, INT_TYPE, 1);
  if(dataSize > 0) {
    data = (char *)malloc(dataSize);
    Receive(connection, data, dataSize);
    if(message != expected || body == NULL)
      free(data);
  }
  if(message != expected)
    return 0;
  if(body != NULL)
    *body = data;
  return 1;

}


/*
 * Sends #message# accompanied by the #dataSize#-long #data# to #host#, then
 * receives the reply and compares it to #expected#.  If the messages match,
 * returns 1 and sets #body# (if it's not NULL) to any data accompanying the
 * incoming message; else returns 0.
 */
static int
Dialogue(NWSC_HostConnection host,
         NwsMessages message,
         const void *data,
         unsigned dataSize,
         NwsMessages expected,
         char **body) {
  char networkHeader[4 + 4 + 4];
  int returnValue;
  HostToNetwork(networkHeader, &NWS_VERSION, INT_TYPE, 1);
  HostToNetwork(networkHeader + 4, &message, INT_TYPE, 1);
  HostToNetwork(networkHeader + 8, &dataSize, INT_TYPE, 1);
  Send(host, networkHeader, sizeof(networkHeader));
  if(dataSize > 0)
    Send(host, data, dataSize);
  returnValue = ReceiveMessage(host, expected, body);
  return returnValue;
}

         
static void
NetworkToForecast(NWSC_Forecast *forecast,
                  const char *networkFormat) {
  NetworkToHost(&forecast->forecast, forecast, DOUBLE_TYPE, 1);
  NetworkToHost(&forecast->error, forecast + 8, DOUBLE_TYPE, 1);
  NetworkToHost(&forecast->methodUsed, forecast + 16, INT_TYPE, 1);
}


int
NWSC_HostConnect(const char *who,
                 short port,
                 NWSC_HostConnection *host) {
  Address addr;
  char *colonPlace;
  char *name = strdup(who);
  int returnValue;
  if((colonPlace = strchr(name, ':')) != NULL) {
    *colonPlace++ = '\0';
    port = atoi(colonPlace);
  }
  returnValue = AddressValue(name, &addr) &&
                (*host = ConnectToPort(addr, port)) != NO_SOCKET;
  free(name);
  return returnValue;
}


void
NWSC_HostDisconnect(NWSC_HostConnection *host) {
  Disconnect(host);
}


int
NWSC_HostGetInfo(NWSC_HostConnection host,
                 NWSC_HostInfo *info) {
  char *data;
  if(!Dialogue(host, HOST_TEST, NULL, 0, HOST_TEST_RESULT, &data))
    return 0;
  memcpy(info->registrationName, data, NWSC_MAX_HOST_NAME);
  NetworkToHost(&info->hostType, data + NWSC_MAX_HOST_NAME, INT_TYPE, 1);
  memcpy(info->nameServer, data + NWSC_MAX_HOST_NAME + 4, NWSC_MAX_HOST_NAME);
  NetworkToHost
    (&info->healthy, data + NWSC_MAX_HOST_NAME + 4 + NWSC_MAX_HOST_NAME,
     INT_TYPE, 1);
  free(data);
  return 1;
}


int
NWSC_HostHalt(NWSC_HostConnection host,
              const char *password) {
  return Dialogue
    (host, HOST_DIE, password,
     (password == NULL) ? 0 : strlen(password) + 1, HOST_DYING, NULL);
}


int
NWSC_HostRegister(NWSC_HostConnection host,
                  const char *who) {
  return Dialogue
    (host, HOST_REGISTER, who, strlen(who) + 1, HOST_REGISTERED, NULL);
}


int
NWSC_HostToggleDiagnostics(NWSC_HostConnection host,
                           int whichDiagnostics) {
  char networkInt[4];
  HostToNetwork(networkInt, &whichDiagnostics, INT_TYPE, 1);
  return Dialogue
    (host, HOST_DIAGNOSTICS, networkInt, sizeof(networkInt),
     HOST_DIAGNOSTICS_ACK, NULL);
}


int
NWSC_ForecasterForecastFromMeasurements(NWSC_HostConnection forecaster,
                                        const NWSC_Measurement *measurements,
                                        unsigned measurementsCount,
                                        NWSC_Forecast *maeForecast,
                                        NWSC_Forecast *mseForecast) {

  int atMost = 1;
  char *data;
  unsigned dataSize;
  int forecastCount;
  int moreToCome = 0;
  char networkDataHeader[4 + 4 + FORECAST_NAME_SIZE + 4];
  char *result;

  HostToNetwork(networkDataHeader, &moreToCome, INT_TYPE, 1);
  HostToNetwork(networkDataHeader + 4, &atMost, INT_TYPE, 1);
  strcpy(networkDataHeader + 8, "");
  HostToNetwork(networkDataHeader + 8 + FORECAST_NAME_SIZE,
                &measurementsCount, INT_TYPE, 1);
  dataSize = sizeof(networkDataHeader) + measurementsCount * 16;
  data = (char *)malloc(dataSize);
  memcpy(data, networkDataHeader, sizeof(networkDataHeader));
  HostToNetwork(data + sizeof(networkDataHeader), measurements, DOUBLE_TYPE,
                measurementsCount * 2);
  if(!Dialogue(forecaster, FORE_DATA, data, dataSize, FORE_FORECAST, &result)) {
    free(data);
    return 0;
  }
  NetworkToHost(&forecastCount, result + FORECAST_NAME_SIZE, INT_TYPE, 1);
  if(forecastCount == 1) {
    NetworkToForecast(maeForecast, result + FORECAST_NAME_SIZE + 4);
    NetworkToForecast(mseForecast, result + FORECAST_NAME_SIZE + 24);
  }
  free(result);
  return forecastCount == 1;

}
                          

int
NWSC_ForecasterMethods(NWSC_HostConnection forecaster,
                       char **methods) {
  return Dialogue(forecaster, METHODS_ASK, NULL, 0, METHODS_TELL, methods);
}


int
NWSC_ForecasterForecastFromSeries(NWSC_HostConnection forecaster,
                                  const char *memoryName,
                                  const char *seriesName,
                                  NWSC_Forecast *maeForecast,
                                  NWSC_Forecast *mseForecast) {

  int atMost = 1;
  int forecastCount;
  char networkSeriesHeader
    [4 + 4 + 4 + FORECAST_NAME_SIZE + NWSC_MAX_HOST_NAME + STATE_NAME_SIZE];
  int moreToCome = 0;
  char *result;
  int seriesCount = 1;

  HostToNetwork(networkSeriesHeader, &seriesCount, INT_TYPE, 1);
  HostToNetwork(networkSeriesHeader + 4, &moreToCome, INT_TYPE, 1);
  HostToNetwork(networkSeriesHeader + 8, &atMost, INT_TYPE, 1);
  strcpy(networkSeriesHeader + 12, seriesName);
  strcpy(networkSeriesHeader + 12 + FORECAST_NAME_SIZE, memoryName);
  strcpy(networkSeriesHeader + 12 + FORECAST_NAME_SIZE + NWSC_MAX_HOST_NAME,
         seriesName);
  if(!Dialogue(forecaster, FORE_SERIES, networkSeriesHeader,
               sizeof(networkSeriesHeader), FORE_FORECAST, &result))
    return 0;
  NetworkToHost(&forecastCount, result + FORECAST_NAME_SIZE, INT_TYPE, 1);
  if(forecastCount == 1) {
    NetworkToForecast(maeForecast, result + FORECAST_NAME_SIZE + 4);
    NetworkToForecast(mseForecast, result + FORECAST_NAME_SIZE + 24);
  }
  free(result);
  return forecastCount == 1;

}


int
NWSC_MemoryAutofetch(NWSC_HostConnection memory,
                     const char *names) {
  return Dialogue
    (memory, AUTOFETCH_BEGIN, names, strlen(names) + 1, AUTOFETCH_ACK, NULL);
}


int
NWSC_MemoryAutofetchCheck(NWSC_HostConnection memory,
                          char **seriesName,
                          char **records,
                          unsigned *recordSize) {
  char *fetched;
  char networkState[STATE_NAME_SIZE + 4 + 4 + 8 + 8];
  if(!CheckIfReadable(memory, 0) ||
     !ReceiveMessage(memory, STATE_FETCHED, &fetched))
    return 0;
  *seriesName = strdup(fetched);
  NetworkToHost(recordSize, fetched + STATE_NAME_SIZE, INT_TYPE, 1);
  *records = strdup(fetched + sizeof(networkState));
  free(fetched);
  return 1;
}


int
NWSC_MemoryClean(NWSC_HostConnection memory,
                 unsigned seconds) {
  char networkInt[4];
  HostToNetwork(networkInt, &seconds, INT_TYPE, 1);
  return Dialogue
    (memory, MEMORY_CLEAN, networkInt, sizeof(networkInt),
     MEMORY_CLEANED, NULL);
}


int
NWSC_MemoryFetch(NWSC_HostConnection memory,
                 const char *seriesName,
                 unsigned howMany,
                 char **records,
                 unsigned *recordSize) {
  char *fetched;
  char networkState[STATE_NAME_SIZE + 4 + 4 + 8 + 8];
  memset(networkState, 0, sizeof(networkState));
  strcpy(networkState, seriesName);
  HostToNetwork(networkState + STATE_NAME_SIZE + 4, &howMany, INT_TYPE, 1);
  if(!Dialogue(memory, FETCH_STATE, networkState, sizeof(networkState),
               STATE_FETCHED, &fetched))
    return 0;
  NetworkToHost(recordSize, fetched + STATE_NAME_SIZE, INT_TYPE, 1);
  *records = strdup(fetched + sizeof(networkState));
  free(fetched);
  return 1;
}


int
NWSC_MemoryStore(NWSC_HostConnection memory,
                 const char *seriesName,
                 const char *records,
                 unsigned recordSize) {

  char *data;
  int dataSize;
  char networkState[STATE_NAME_SIZE + 4 + 4 + 8 + 8];
  double now = (int)time(NULL);
  int recordCount = strlen(records) / recordSize;
  int returnValue;
  double tenYears = 315360000.0;

  memset(networkState, 0, sizeof(networkState));
  strcpy(networkState, seriesName);
  HostToNetwork(networkState + STATE_NAME_SIZE, &recordSize, INT_TYPE, 1);
  HostToNetwork(networkState + STATE_NAME_SIZE + 4, &recordCount, INT_TYPE, 1);
  HostToNetwork(networkState + STATE_NAME_SIZE + 8, &now, DOUBLE_TYPE, 1);
  HostToNetwork(networkState + STATE_NAME_SIZE + 16, &tenYears, DOUBLE_TYPE, 1);
  dataSize = sizeof(networkState) + strlen(records) + 1;
  data = (char *)malloc(dataSize);
  memcpy(data, networkState, sizeof(networkState));
  strcpy(data + sizeof(networkState), records);
  returnValue =
    Dialogue(memory, STORE_STATE, data, dataSize, STATE_SORTED, NULL);
  free(data);
  return returnValue;

}


int
NWSC_NameServerRegister(NWSC_HostConnection nameServer,
                        const char *object,
                        unsigned forHowLong) {
  char *data;
  int objectSize = strlen(object) + 1;
  int returnValue;
  data = (char *)malloc(objectSize + 4);
  memcpy(data, object, objectSize);
  HostToNetwork(data + objectSize, &forHowLong, INT_TYPE, 1);
  returnValue =
    Dialogue(nameServer, NS_REGISTER, data, objectSize+4, NS_REGISTERED, NULL);
  free(data);
  return returnValue;
}


int
NWSC_NameServerSearch(NWSC_HostConnection nameServer,
                      const char *filter,
                      char **whereTo) {
  return Dialogue
    (nameServer, NS_SEARCH, filter, strlen(filter) + 1, NS_SEARCHED, whereTo);
}


int
NWSC_NameServerUnregister(NWSC_HostConnection nameServer,
                          const char *filter) {
  return Dialogue
    (nameServer, NS_UNREGISTER, filter, strlen(filter) + 1,
     NS_UNREGISTERED, NULL);
}


int
NWSC_SensorActivityStart(NWSC_HostConnection sensor,
                         const char *activityName,
                         const char *control,
                         const char *skill,
                         const char *options) {
  char *data;
  unsigned dataSize;
  int returnValue;
  dataSize = strlen(activityName) + 1 + strlen(control) + 1 +strlen(skill) +
             1 + strlen(options) + 1;
  data = (char *)malloc(dataSize);
  strcpy(data, activityName);
  strcpy(data + strlen(activityName) + 1, control);
  strcpy(data + strlen(activityName) + 1 + strlen(control) + 1, skill);
  strcpy
    (data + strlen(activityName) + 1 + strlen(control) + 1 + strlen(skill) + 1,
     options);
  returnValue =
    Dialogue(sensor, ACTIVITY_START, data, dataSize, ACTIVITY_STARTED, NULL);
  free(data);
  return returnValue;
}


int
NWSC_SensorActivityStop(NWSC_HostConnection sensor,
                        const char *activityName) {
  return Dialogue
    (sensor, ACTIVITY_STOP, activityName, strlen(activityName) + 1,
     ACTIVITY_STOPPED, NULL);
}


int
NWSC_SensorTcpConnectionTest(NWSC_HostConnection sensor,
                             unsigned *result) {
  double before;
  Socket connection;
  before = time(NULL);
  connection = ConnectToPort(PeerAddress(sensor), PeerPort(sensor));
  *result = (unsigned)((time(NULL) - before) * 1000.0);
  Disconnect(&connection);
  return connection != NO_SOCKET;
}


int
NWSC_SensorTcpMessageTest(NWSC_HostConnection sensor,
                          unsigned experimentSize,
                          unsigned bufferSize,
                          unsigned messageSize,
                          double *bandwidth,
                          double *latency) {

  char byte[1] = "";
  Socket connection;
  char *data;
  unsigned leftToDo;
  char *message;
  char networkTcpControl[4 + 4 + 4];
  short port;
  int returnValue;
  double startTime;

  HostToNetwork(networkTcpControl, &experimentSize, INT_TYPE, 1);
  HostToNetwork(networkTcpControl + 4, &bufferSize, INT_TYPE, 1);
  HostToNetwork(networkTcpControl + 8, &messageSize, INT_TYPE, 1);
  if(!Dialogue(sensor, TCP_BW_REQ, networkTcpControl, sizeof(networkTcpControl),
               TCP_HANDSHAKE, &data))
    return 0;

  NetworkToHost(&port, data + 4, SHORT_TYPE, 1);
  free(data);
  if(port == 0)
    connection = sensor;
  else if((connection = ConnectToPortBuffered(PeerAddress(sensor), port,
                                              bufferSize, bufferSize)) ==
          NO_SOCKET)
    return 0;

  startTime = time(NULL);
  if(!Send(connection, byte, 1) || !Receive(connection, byte, 1)) {
    if(connection != sensor)
      Disconnect(&connection);
    return 0;
  }
  *latency = (time(NULL) - startTime) * 1000.0;

  message = (char *)malloc(messageSize);
  memset(message, 0, messageSize);
  startTime = time(NULL);
  for(leftToDo = experimentSize ; leftToDo > 0; leftToDo -= messageSize)
    Send(connection, message, messageSize);
  returnValue = CheckIfReadable(connection, 10.0);
  *bandwidth = (experimentSize*8.0/1000000.0) / (time(NULL)-startTime);
  free(message);
  if(connection != sensor)
    Disconnect(&connection);
  return returnValue;

}
