#! /usr/bin/perl -w
# $Id: rasha.pl,v 1.25 2004/04/03 22:24:36 jhayes Exp $

my $USAGE = "rasha [-d] [host[:port]]";

# This is the rash agent that runs on a resource not normally accessible from
# outside.  The host:port command-line argument gives contact information for
# the rash daemon; if absent, it defaults to localhost:7000.  See rashd for a
# description of the protocol.
#
# The command portion of the RUN message supported by rasha has the format
# [preamble ';' ...] path [switch ...].  Each preamble is either a working
# directory specification ('cd' [path]), an environment variable setting ('set'
# name [word]), or a redirection ('redirect' fd path).  The default working
# directory path is the home directory.  Environment variable values may be
# enclosed in either single or double quotes, and escaping with backslash is
# supported.  A command can also contain environment variable references of the
# form '${' name '}'.  To provide some minimal security, the agent restricts
# the programs it runs to those in its working directory and subdirectories,
# plus a few in (/usr)/bin known to be safe.

use Cwd;
use IO::Handle;
use IO::Select;
use IO::Socket;

my $CRLF = "\015\012";
my $buffer = '';
my $debug;
my %exitCodes;
my $name;
my $pid;
my %pids;
my $pipeName = 'PIPE0';
my $sock;
my %stderrs;
my %stdins;
my %stdouts;
my $wordPat = q#'((\\.|[^'])*)'|"((\\.|[^"])*)"|([^'";\s]+)#;

$debug = shift if $#ARGV >= 0 && $ARGV[0] eq '-d';
push @ARGV, 'localhost' if $#ARGV < 0;
die "$USAGE" if $#ARGV != 0 || $ARGV[0] !~ /^([^:]+)(:(\d+))?$/;
$sock = IO::Socket->new
  (Domain => AF_INET, PeerAddr => $1, PeerPort => defined $3 ? $3 : 7000)
  or die "socket: $!";
$sock->autoflush(1);
$SIG{PIPE} = 'IGNORE';

if(!defined $debug) {
  open STDERR, '/dev/null';
  open STDOUT, '/dev/null';
}

# Splits #text# into a list of quoted and unquoted words and returns the
# result.  If #strip# is non-zero, quoted words are stripped of quotes and
# escaped characters.
sub GetWords {
  my ($text, $strip) = @_;
  my @result;
  while($text =~ s/\s*($wordPat)//) {
    push(@result, $1) and next if !$strip || defined $6;
    my $word = defined $2 ? $2 : $4;
    $word =~ s/\\(.)/$1/g;
    push(@result, $word);
  }
  return @result;
}

# Returns a message received on #from#, minus the CRLF terminator and with
# special characters decoded.
sub Receive {
  my ($from) = @_;
  my $buf = ' ';
  my $msg = '';
  $msg .= $buf while $msg !~ s/$CRLF$// && sysread($from, $buf, 1);
  $msg =~ s/\\n/\n/g;
  $msg =~ s/\\\\/\\/g;
  return $msg =~ /./ ? $msg : undef;
}

# Sends #msg#, followed by CRLF and with special characters encoded, to #to#.
sub Send {
  my ($to, $msg) = @_;
  $msg =~ s/\\/\\\\/g;
  $msg =~ s/\n/\\n/g;
  print STDOUT "$msg\n";
  $to->print("$msg$CRLF");
}

while(1) {
  IO::Select->new($sock, values %stderrs, values %stdouts)->can_read();
  while(IO::Select->new($sock)->can_read(0)) {
    my $msg = &Receive($sock);
    if(!defined $msg) {
      kill KILL, values %pids;
      die "Lost connection\n";
    }
    print STDOUT "Received '$msg'\n";
    if($msg =~ /^KILL (\S+)/) {
      kill(KILL, $pids{$1}) if defined $pids{$1};
    } elsif($msg =~ s/^RUN \S+ (\S+) //) {
      my ($errRead, $errWrite, $exitRead, $exitWrite, $inRead, $inWrite,
          $outRead, $outWrite);
      $name = $1;
      pipe($errRead = ++$pipeName, $errWrite = ++$pipeName);
      pipe($exitRead = ++$pipeName, $exitWrite = ++$pipeName);
      pipe($inRead = ++$pipeName, $inWrite = ++$pipeName);
      pipe($outRead = ++$pipeName, $outWrite = ++$pipeName);
      if(($pid = fork()) == 0) {
        # The blocking can_read() above will remain blocked when the command
        # exits if both stderr and stdout are redirected.  To handle this case,
        # we interpose another process that holds the pipes open, waits for the
        # command to exit, and forwards the exit code.
        if(($pid = fork()) > 0) {
          wait();
          exit $? >> 8;
        }
        open(STDERR, ">&$errWrite");
        open(STDIN,  "<&$inRead");
        open(STDOUT, ">&$outWrite");
        close $errRead;
        close $errWrite;
        close $exitRead;
        close $exitWrite;
        close $inRead;
        close $inWrite;
        close $outRead;
        close $outWrite;
        while($msg =~ s/^((($wordPat)\s*)+)(;\s*|$)//) {
          my $word;
          my @words = &GetWords($1, 0);
          foreach $word (@words) {
            next if $word =~ /^'/;
            $word = $1 . (defined $ENV{$2} ? $ENV{$2} : "") . $3
              while $word =~ /^(.*)\${([^}]+)}(.*)$/;
          }
          @words = &GetWords(join(' ', @words), 1);
          if($words[0] eq 'cd' && defined $words[1]) {
            chdir $words[1] or (print STDERR "cd failed\n" and exit 1);
          } elsif($words[0] eq 'cd') {
            chdir or (print STDERR "cd failed\n" and exit 1);
          } elsif($words[0] eq 'set' and defined $words[1]) {
            if(defined $words[2]) {
              $ENV{$words[1]} = $words[2];
            } else {
              delete $ENV{$words[1]};
            }
          } elsif($words[0] eq 'redirect') {
            print STDERR "Missing fd\n" and exit 1 if !defined $words[1];
            print STDERR "Bad fd\n" and exit 1 if $words[1] !~ /^\d+$/;
            print STDERR "Missing path\n" and exit 1 if !defined $words[2];
            open(STDIN, "<$words[2]") or
              (print STDERR "stdin failed\n" and exit 1) if $words[1] eq '0';
            open(STDOUT, ">$words[2]") or
              (print STDERR "stdout failed\n" and exit 1) if $words[1] eq '1';
            open(STDERR, ">$words[2]") or
              (print STDERR "stderr failed\n" and exit 1) if $words[1] eq '2';
          } else {
            autoflush STDERR 1;
            autoflush STDOUT 1;
            if($words[0] eq '??') {
              foreach $name (sort keys %pids) {
                print STDOUT "$name ($pids{$name})" .
                  (defined $exitCodes{$name} ? " [$exitCodes{$name}]\n" : "\n");
              }
              exit 0;
            }
            # Restrict execs to rasha wd & subdirs, plus a few known to be safe.
            if($words[0] !~ m#/#) {
              foreach $dir (split(':', $ENV{"PATH"})) {
                $words[0] = "$dir/$words[0]" and last
                  if -x "$dir/$words[0]" || -x "$dir/$words[0].exe";
              }
            }
            while($words[0] =~ s#(^|/)\./#$1#) { }
            while($words[0] =~ s#[^/]+/\.\./##) { }
            my $cwd = cwd();
            $words[0] = "$cwd/$words[0]" if $words[0] !~ m#^/#;
            print STDERR "Permission denied\n" and exit 1
              if $words[0] !~ m#^(/usr)?/bin/(date|df|false|hostname|ls|sleep|test|touch|true|pwd|uname|uptime|vmstat|w|wc|who)(\.exe)?$# &&
                 $words[0] !~ m#^$cwd#;
            print STDERR "Permission denied\n" and exit 1
              if ! -x $words[0] && ! -x "$words[0].exe";
            exec {$words[0]} @words or
              (print STDERR "exec failed\n" and exit 88);
          }
        }
        exit 0; # No application to exec, just useless preamble.
      }
      $stderrs{$name} = IO::Handle->new_from_fd(fileno($errRead), "r");
      $stdins{$name} = IO::Handle->new_from_fd(fileno($inWrite), "w");
      $stdouts{$name} = IO::Handle->new_from_fd(fileno($outRead), "r");
      $stdins{$name}->autoflush(1);
      $pids{$name} = $pid;
      close $errWrite;
      close $exitWrite;
      close $inRead;
      close $outWrite;
    } elsif($msg =~ s/STDIN (\S+) //) {
      $stdins{$1}->print($msg) if defined $stdins{$1};
    }
  }
  foreach $name (keys %stderrs) {
    next if !IO::Select->new($stderrs{$name})->can_read(0);
    if($stderrs{$name}->sysread($buffer, 1024) > 0) {
      print STDERR "Read from stderrs{$name}\n";
      &Send($sock, "STDERR $name $buffer");
    } else {
      print STDERR "Close stderrs{$name}\n";
      $stderrs{$name}->close();
      delete $stderrs{$name};
      # Drain stdout pipe before forwarding exit code.
      &Send($sock, "STDOUT $name $buffer")
        while defined $stdouts{$name} &&
        $stdouts{$name}->sysread($buffer, 1024) > 0;
      $stdins{$name}->close();
      delete $stdins{$name};
      waitpid($pids{$name}, 0);
      $exitCodes{$name} = $?;
      &Send($sock, "FINISHED $name ". ($exitCodes{$name} >> 8));
    }
  }
  foreach $name (keys %stdouts) {
    next if !IO::Select->new($stdouts{$name})->can_read(0);
    if($stdouts{$name}->sysread($buffer, 1024) > 0) {
      print STDERR "Read from stdouts{$name}\n";
      &Send($sock, "STDOUT $name $buffer");
    } else {
      print STDERR "Close stdouts{$name}\n";
      $stdouts{$name}->close();
      delete $stdouts{$name};
    }
  }
}

$sock->close();
