/* $Id: nwsclient.h,v 1.4 2004/01/12 19:11:27 jhayes Exp $ */
/*
 * Copyright  2004 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#ifndef NWSCLIENT_H
#define NWSCLIENT_H


/*
 * This package defines a client interface to Network Weather Service hosts.
 * This interface is much lower-level than the NWS sources' nws_api.h; the
 * functions below mostly map very closely to the NWS protocol messages.  This
 * allows client code much greater control over the use of the NWS services, at
 * the expense of requiring more intricate knowledge of the NWS protocol.
 *
 * Presently unsupported: clique protocol messages; forecaster history (the
 * NWSC_ForecastFrom* functions return a single forecast pair) and forecaster
 * requests for multiple series in a single message.
 *
 * All the functions other than NWSC_HostDisconnect return 1 if successful,
 * else 0.  Character strings returned by the functions are malloc'ed and
 * should be freed after use.
 */


#ifdef __cplusplus
extern "C" {
#endif


/* Default ports for the various NWS hosts. */
#define NWSC_FORECASTER_PORT 8080
#define NWSC_MEMORY_PORT 8050
#define NWSC_NAME_SERVER_PORT 8090
#define NWSC_SENSOR_PORT 8060

/* The arbitrary limits the NWS places on host name size. */
#define NWSC_MAX_MACHINE_NAME 64
#define NWSC_MAX_PORT_IMAGE 5
#define NWSC_MAX_HOST_NAME (NWSC_MAX_MACHINE_NAME + 1 + NWSC_MAX_PORT_IMAGE)

/* A connection to an NWS host (in reality, a socket). */
typedef int NWSC_HostConnection;

/* Types of NWS hosts. */
typedef unsigned NWSC_HostType;
#define NWSC_FORECASTER_HOST 0
#define NWSC_MEMORY_HOST 1
#define NWSC_NAME_SERVER_HOST 2
#define NWSC_SENSOR_HOST 3

/* Types of NWS diagnostics messages. */
#define NWSC_ALL_ERRORS -1
#define NWSC_ALL_LOGS -2
#define NWSC_ALL_DIAGNOSTICS -3

/* Information returned by NWSC_HostGetInfo */
typedef struct {
  char registrationName[NWSC_MAX_HOST_NAME];
  NWSC_HostType hostType;
  char nameServer[NWSC_MAX_HOST_NAME];
  int healthy;
} NWSC_HostInfo;

/* An NWS resource measurement. */
typedef struct {
  double timeStamp;
  double measurement;
} NWSC_Measurement;

/* An NWS forecast of resource availability. */
typedef struct {
  double forecast;
  double error;
  unsigned methodUsed;
} NWSC_Forecast;


/*
 * Returns in #connection# a connection to the NWS host specified by #who#,
 * which has the format dnsName[:port] or ipAddress[:port].  #port# is used in
 * making the connection if no port is specified as part of #host#.
 */
int
NWSC_HostConnect(const char *who,
                 short port,
                 NWSC_HostConnection *host);


/* Closes the NWS host connection #host#. */
void
NWSC_HostDisconnect(NWSC_HostConnection *host);


/* Returns in #info# host information for #host#. */
int
NWSC_HostGetInfo(NWSC_HostConnection host,
                 NWSC_HostInfo *info);


/*
 * Requests that #host# terminate, passing #password# for verification if it is
 * not NULL.
 */
int
NWSC_HostHalt(NWSC_HostConnection host,
              const char *password);


/*
 * Directs #host# to change its name server to the host specified by #who#,
 * which has the format dnsName[:port] or ipAddress[:port].
 */
int
NWSC_HostRegister(NWSC_HostConnection host,
                  const char *who);


/*
 * Directs #host# to toggle production of #whichOnes# (which should be one of
 * the constants defined above) diagnostics messages.  If the host is presently
 * writing these diagnostics, they will be suppressed after the call; if the
 * host is presently suppressing them, they will be written after the call.
 */
int
NWSC_HostToggleDiagnostics(NWSC_HostConnection host,
                           int whichDiagnostics);


/*
 * Sends the #measurementsCount#-long list #measurements# to #forecaster# and
 * returns the best mean average error and mean square error forecasts in
 * #maeForecast# and #mseForecast#.
 */
int
NWSC_ForecasterForecastFromMeasurements(NWSC_HostConnection forecaster,
                                        const NWSC_Measurement *measurements,
                                        unsigned measurementsCount,
                                        NWSC_Forecast *maeForecast,
                                        NWSC_Forecast *mseForecast);


/*
 * Asks #forecaster# to generate forecasts for #seriesName#.  #memory#
 * specifies the memory that has the series measurements; it has the format
 * dnsName[:port] or ipAddress[:port].  The returned forecasts are returned
 * int #maeForecast# and #mseForecast#.
 */
int
NWSC_ForecasterForecastFromSeries(NWSC_HostConnection forecaster,
                                  const char *memory,
                                  const char *seriesName,
                                  NWSC_Forecast *maeForecast,
                                  NWSC_Forecast *mseForecast);


/*
 * Returns in #methods#, as a tab-delimited list, the forecasting methods used
 * by #forecaster#.
 */
int
NWSC_ForecasterMethods(NWSC_HostConnection forecaster,
                       char **methods);


/*
 * Directs #memory# to delete all series that have not been changed within the
 * past #seconds# seconds.
 */
int
NWSC_MemoryClean(NWSC_HostConnection memory,
                 unsigned seconds);


/*
 * Directs #memory# to begin automatically forwarding records stored in any of
 * the series named in the tab-delimited list #names#.  This overrides any
 * previous auto-fetch request.
 */
int
NWSC_MemoryAutofetch(NWSC_HostConnection memory,
                     const char *names);


/*
 * Checks to see if #memory# has forwarded any stored records.  If so, returns
 * the name of the series in which records were stored in #seriesName#, the
 * records themselves in #records#, and the record length in #recordSize#.
 * (NWS records are fixed length with no delimiter.)
 */
int
NWSC_MemoryAutofetchCheck(NWSC_HostConnection memory,
                          char **seriesName,
                          char **records,
                          unsigned *recordSize);


/*
 * Retrieves from #memory# the #howMany# most recently stored records from
 * series #seriesName#.  Returns the records in #records# and the record length
 * in #recordSize#.
 */
int
NWSC_MemoryFetch(NWSC_HostConnection memory,
                 const char *seriesName,
                 unsigned howMany,
                 char **records,
                 unsigned *recordSize);


/*
 * Directs #memory# to store in series #seriesName# the information contained
 * in #records#, which is divided into #recordSize#-byte records.
 */
int
NWSC_MemoryStore(NWSC_HostConnection memory,
                 const char *seriesName,
                 const char *records,
                 unsigned recordSize);


/*
 * Directs #nameServer# to register an object consisting of the set of
 * attributes in #object#, to expire after #forHowLong# seconds.  #object# is
 * a tab-delimited and -terminated list of name:value pairs.
 */
int
NWSC_NameServerRegister(NWSC_HostConnection nameServer,
                        const char *object,
                        unsigned forHowLong);


/*
 * Returns in #whereTo# all objects registered with #nameServer# that match
 * #filter#, an LDAP-style search filter.  Each object in the returned value is
 * a tab-delimited list of name:value pairs, and multiple objects are separated
 * by pairs of tabs.
 */
int
NWSC_NameServerSearch(NWSC_HostConnection nameServer,
                      const char *filter,
                      char **whereTo);


/*
 * Directs #nameServer# to delete the registration for all objects that match
 * #filter#. 
 */
int
NWSC_NameServerUnregister(NWSC_HostConnection nameServer,
                          const char *filter);


/*
 * Directs #sensor# to begin a new activity and register it as #activityName#.
 * #control#, #skill#, and #options# configure the activity.
 */
int
NWSC_SensorActivityStart(NWSC_HostConnection sensor,
                         const char *activityName,
                         const char *control,
                         const char *skill,
                         const char *options);


/* Directs #sensor# to terminate the activity registered as #activityName. */
int
NWSC_SensorActivityStop(NWSC_HostConnection sensor,
                        const char *activityName);


/*
 * Returns in #result# the number of milliseconds required to establish a new
 * connection to #sensor#.
 */
int
NWSC_SensorTcpConnectionTest(NWSC_HostConnection sensor,
                             unsigned *result);


/*
 * Returns in #bandwidth# and #latency# the bandwidth (Mb/sec) and latency
 * (milliseconds) to #sensor#.  Configures the bandwidth test as specified by
 * #experimentSize#, #bufferSize#, and #messageSize#.
 */
int
NWSC_SensorTcpMessageTest(NWSC_HostConnection sensor,
                          unsigned experimentSize,
                          unsigned bufferSize,
                          unsigned messageSize,
                          double *bandwidth,
                          double *latency);


#ifdef __cplusplus
}
#endif


#endif
