/* $Id: corestructures.h,v 1.64 2003/10/02 15:43:26 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#ifndef CORESTRUCTURES_H
#define CORESTRUCTURES_H

/*
 * Core data structures and function definitions for a master/worker scheduler.
 * The scheduler should attempt to work even with not all the fields filled out
 * (e.g. if host IP address is not given, it should try to look it up using
 * name). In fact, it is intended that all the MetaData fields are updated only
 * by the scheduler, relying on NWS and other information sources.
 */

#include <hashseed.h>
#include <elmetric.h>
#include "elbase.h"

/* Various hash table sizes. */
#define LARGE_HASH  2023
#define LINKED_LIST 1
#define MEDIUM_HASH 517
#define SMALL_HASH  13

/* defining the list types, which are really hash-tables */
typedef ASH_HashTable DiskList;
typedef ASH_HashTable FileCopyList;
typedef ASH_HashTable FileList;
typedef ASH_HashTable FileTransferList;
typedef ASH_HashTable GridInfoSourceList;
typedef ASH_HashTable HostList;
typedef ASH_HashTable TaskList;

typedef struct _Disk           *Disk;
typedef struct _FileCopy       *FileCopy;
typedef struct _File           *File;
typedef struct _FileTransfer   *FileTransfer;
typedef struct _GridInfoSource *GridInfoSource;
typedef struct _Host           *Host;
typedef struct _Task           *Task;

/* Number unique to a particular instatiation of an object. */
typedef unsigned long SequenceNumber;

/* Statii of actions like tasks and file transfers. */
typedef enum {
  ACTION_CANCELED,    /* the action was canceled          */
  ACTION_DONE,        /* the action succesfully completed */
  ACTION_FAILED,      /* the action failed                */
  ACTION_IN_PROGRESS, /* the action has been started      */
  ACTION_NEW          /* this action has not been started */
} ActionStatuses;


/**** DISK ****/

struct _Disk {
  char *id;      /* unique name of this disk */
  char *datadir; /* path where files are found on the disk */
  ELBASE_StorageServiceTypes method; /* access method for this disk */
  char *server;  /* method-specific contact string for this disk */
  FileTransfer *transfersCompleted;  /* vector of transfers completed */
  FileTransfer *transfersInProgress; /* vector of transfers in-progress */
};

void
Disk_Free(Disk d);

Disk
Disk_New(char *id,
         char *datadir,
         ELBASE_StorageServiceTypes method,
         char *server);

void
DiskList_Add(DiskList dl,
             Disk d);

void
DiskList_Free(DiskList dl,
              int freeElements);

#define DiskList_Get(l,s) ((Disk)ASH_GetValue(l, s, NULL))
#define DiskList_New(n) \
  ASH_HashTableNew(n, &ASH_StringCompare, &ASH_StringHash, NULL)


/**** FILE ****/

struct _File {
  char *name;          /* unique file name */
  double size;         /* file size in bytes; 0 if unknown */
  Task outputOf;       /* (a/the) task that produces this file */
  int download;        /* is this output file to be copied to local disk? */
  int transferable;    /* may this file be transferred between disks? */
  FileCopyList copies; /* existing copies of this file */
};

void
File_Free(File f);

File
File_New(char *name);

void
FileList_Add(FileList fl,
             File f);

void
FileList_Free(FileList fl,
              int freeElements);

#define FileList_Get(l,s) ((File)ASH_GetValue(l, s, NULL))
#define FileList_New(n) \
  ASH_HashTableNew(n, &ASH_StringCompare, &ASH_StringHash, NULL)


/**** FILE COPY ****/

struct _FileCopy {
  Disk disk;        /* disk where this copy resides */
  char *path;       /* local file path for this copy */
};

void
FileCopy_Free(FileCopy fc);

FileCopy
FileCopy_New(Disk disk,
             char *path);

void
FileCopyList_Add(FileCopyList fcl,
                 FileCopy fc);

void
FileCopyList_Free(FileCopyList fcl,
                  int freeElements);

#define FileCopyList_Get(l,s) ((FileCopy)ASH_GetValue(l, s, NULL))
#define FileCopyList_New(n) \
  ASH_HashTableNew(n, &ASH_StringCompare, &ASH_StringHash, NULL)


/**** FILE TRANSFER ****/

typedef struct _FileTransferMetaData {
  long timeStarted;   /* time when transfer began */
  long timeCompleted; /* time when transfer finished */
  long timeEstimated; /* estimated transfer time */
  double bandwidth;   /* estimated bandwidth on link */
  double latency;     /* estimated latency on link */
} FileTransferMetaData;

struct _FileTransfer {
  char *id;                  /* unique id for this transfer */
  File file;                 /* file being transferred */
  Disk source;               /* source disk */
  Disk dest;                 /* destination disk */
  ActionStatuses stat;       /* transfer status */
  FileTransferMetaData meta; /* meta-data about the transfer */
};

void
FileTransfer_Free(FileTransfer ft);

FileTransfer
FileTransfer_New(File file,
                 Disk source,
                 Disk dest);

void
FileTransferList_Add(FileTransferList ftl,
                     FileTransfer ft);

void
FileTransferList_Free(FileTransferList ftl,
                      int freeElements);

#define FileTransferList_Get(l,s) ((FileTransfer)ASH_GetValue(l, s, NULL))
#define FileTransferList_New(n) \
  ASH_HashTableNew(n, &ASH_StringCompare, &ASH_StringHash, NULL)
#define FileTransferList_Remove(l,ft) ASH_RemoveMapping(l, ft->id, NULL)


/**** GRIDINFO ***/

struct _GridInfoSource {
  char *id;
  ELMETRIC_MetricServiceTypes method;
  char *server;
};

void
GridInfoSource_Free(GridInfoSource gis);

GridInfoSource
GridInfoSource_New(char *id,
                   ELMETRIC_MetricServiceTypes method,
                   char *server);

void
GridInfoSourceList_Add(GridInfoSourceList gisl,
                       GridInfoSource gis);

void
GridInfoSourceList_Free(GridInfoSourceList gisl,
                        int freeElements);

#define GridInfoSourceList_Get(l,s) ((GridInfoSource)ASH_GetValue(l, s, NULL))
#define GridInfoSourceList_New(n) \
  ASH_HashTableNew(n, &ASH_StringCompare, &ASH_StringHash, NULL)


/**** HOSTS ****/

/**** added by Ming ****/
typedef struct _HostParameters {
  double utilRate;  /* benchmark value for comparing hosts  */
  double arriRate;  /* benchmark value for comparing hosts  */
  double servStd;  /* benchmark value for comparing hosts  */
  double speedRate;  /* benchmark value for comparing hosts  */
} HostParameters;
/**** added by Ming ****/

typedef struct _HostMetaData {
  double benchmark;  /* benchmark value for comparing hosts  */
  long timeDisabled; /* -1: user disabled; 0: enabled; other: failure time */
} HostMetaData;
typedef struct _HostScheduler {
  ELBASE_SchedulerServiceTypes method;
  unsigned port;
  char **attrs;
  unsigned long agentCount;
  struct _HostScheduler *next;
} *HostScheduler;

struct _Host {
  char *id;           /* unique name of this host */
  Disk disk;          /* preferred disk for host */
  char *dnsName;      /* network name of host */
  char *wd;           /* working directory when running on host */
  ELBASE_ComputeServiceTypes method; /* access method for host */
  char *server;       /* method-specific contact string for host */
  unsigned cpus;      /* how many CPUs host has */
  double memory;      /* host memory size in bytes; 0 if unknown */
  Task *tasksCompleted;  /* vector of tasks completed by host */
  Task *tasksInProgress; /* list of tasks running on host */
  HostScheduler schedulers; /* List of scheduler info; NULL if SHELL only */
  HostMetaData meta;  /* meta-data about the host */
};

void
Host_Free(Host h);

Host
Host_New(char *id,
         Disk disk,
         char *wd,
         ELBASE_ComputeServiceTypes method,
         char *server);

void
HostList_Add(HostList hl,
             Host h);

void
HostList_Free(HostList hl,
              int freeElements);

#define HostList_Get(l,s) ((Host)ASH_GetValue(l, s, NULL))
#define HostList_New(n) \
  ASH_HashTableNew(n, &ASH_StringCompare, &ASH_StringHash, NULL)

/**** TASKS ****/

typedef struct _TaskMetaData {
  long timeStarted;   /* time when task began */
  long timeCompleted; /* time when task finished */
  long timeEstimated; /* predicted execution time */
  double cost;        /* user's estimated relative cost computation */
  double memory;      /* user's estimated memory requirement (bytes) */
  double cpuShare;    /* estimated %CPU available at predict time */
} TaskMetaData;

struct _Task {
  SequenceNumber seq;
  unsigned priority;   /* user-assigned task priority, higher gets done first */
  File exec;           /* program to schedule */
  char **arguments;    /* command-line arguments to pass to the executable */
  File *input;         /* vector of input files for the task */
  File *output;        /* vector of files produced by the task */
  ActionStatuses stat; /* status of this task */
  Host host;           /* host it has been scheduled on (initially NULL) */
  TaskMetaData meta;   /* meta-data about the task */
};


/*
 * Sorts the #len#-long array #tasks#.  Sorts by decreasing priority if
 * #byPriority# is non-zero, otherwise by increasing sequence number.
 */
void
SortTasks(Task *tasks,
          unsigned len,
          int byPriority);

void
Task_Free(Task w);

Task
Task_New(void);

void
TaskList_Add(TaskList tl,
             Task w);

void
TaskList_Free(TaskList tl,
              int freeElements);

#define TaskList_Get(l,seq) \
  ((Task)ASH_GetValue(l, (ASH_KeyType)seq, NULL))
#define TaskList_New(n) ASH_HashTableNew(n, NULL, NULL, NULL)
#define TaskList_Remove(l,t) \
  ASH_RemoveMapping(l, (ASH_KeyType)t->seq, NULL)


#endif
