/* $Id: scheduler.c,v 1.101 2003/12/02 23:07:14 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <elmetric.h>
#include <limits.h> /* ULONG_MAX */
#include <stdio.h>  /* sprintf */
#include <string.h> /* string functions */
#include <time.h>   /* time */
#define ASH_SHORT_NAMES
#include <hashseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#define ASXML_SHORT_NAMES
#include <xmlseed.h>
#include "compute.h"
#include "corestructures.h"
#include "gantt.h"
#include "global.h"
#include "misc.h"
#include "predictor.h"
#include "storage.h"
#include "scheduler.h"
#include "ghs.h" //Added by Ming

/*
 * ssh limits the number of concurrent pending connections (see MaxStartups in
 * the sshd_config man page).  When dealing with long latencies, it's easy to
 * reach this number of multi-cpu hosts.  To hack around this, we limit the
 * number of attempted concurrent starts on a single host in sched_run.
 */
#define MAX_CONCURRENT_HOST_STARTS 1
#define MAX_TASKS 40 /* Task-space reduction upper bound */
int scheduleFrequency=900;

extern void GhsMetataskSched(double *tasks, int taskNum, HostParameters *hosts, int hostNum, int *maps);
extern int GetParasValue(double *utils, double *arris, double *servstds, double *speeds, int hostNum, Host *hosts);

//extern (static char *) GetMetricValue(ELMETRIC_MetricTypes metric, const char *entity, Host h);

/* Types of events in our Gantt charts. */
enum {TASK_EVENT, INPUT_TRANSFER_EVENT, OUTPUT_TRANSFER_EVENT};

/* Reasons a task may not be ready for scheduling. */
typedef enum {
  SCHEDULABLE = 0, HOST_DISABLED, INPUT_INACCESSIBLE, INPUT_UNAVAILABLE,
  INSUFFICIENT_MEMORY, LOW_PRIORITY, NOT_NEW
} TaskSchedulability;

/* A single CPU on a host. */
typedef struct _Processor {
  Host host;
  unsigned cpu;
} *Processor;

/* A file transfer link between two disks. */
typedef struct _Link {
  Disk d1, d2;
} *Link;


static GanttChart linkChart = NULL;
static Disk localDisk = NULL;
static GanttChart processorChart = NULL;
static unsigned long expires = 0;
static GanttEvent *scheduledInputTransferEvents = NULL;
/*
 * A boolean that indicates whether output files are transferred as soon as the
 * task is completed, or only after input for subsequent tasks is transferred.
 */
static const int transferOutputImmediately = 0;


/* Returns 1 iff a copy of file #f# exists on disk #d#. */
#define IsFileOnDisk(f,d) (FileCopyList_Get((f)->copies, (d)->id) != NULL)
/* Returns 1 iff #f# is a special file that exists everywhere w/out copying. */
#define IsSpecialFile(f) \
  (*(f)->name == '/' && strncmp((f)->name, "/dev/", 5) == 0)


static unsigned long
PlanTask(Task t,
         Processor p,
         int enterInChart,
         int scheduleOutput);


/* Computes and returns #t# schedulability status. */
static TaskSchedulability
CheckTaskSchedulability(Task t,
                        unsigned priority) {
  unsigned i;
  HashIterator it;
  if(t->priority < priority)
    return LOW_PRIORITY;
  if(t->host != NULL && t->host->meta.timeDisabled != 0)
    return HOST_DISABLED;
  for(i = 0; i < Size(t->input); i++) {
    File f = t->input[i];
    if(GetMappingCount(f->copies) == 0)
      return INPUT_UNAVAILABLE;
    else if(f->transferable)
      continue;
    ForEachMapping(global.hosts, it) {
      Host h = (Host)IteratorValue(it);
      if(h->meta.timeDisabled == 0 && IsFileOnDisk(f, h->disk))
        break;
    }
    if(!it.valid)
      return INPUT_INACCESSIBLE;
  }
  if(t->meta.memory > 0) {
    ForEachMapping(global.hosts, it) {
      Host h = (Host)IteratorValue(it);
      if(h->meta.timeDisabled == 0 && (t->host == NULL || t->host == h) &&
         (h->memory == 0 || h->memory >= t->meta.memory))
        break;
    }
    if(!it.valid)
      return INSUFFICIENT_MEMORY;
  }
  return SCHEDULABLE;
}


/* Returns 1 iff the executable file #f# has a copy that can be run on #h#. */
static int
CompatableExecutable(File f,
                     Host h) {
  return GetMappingCount(f->copies) == 0 ||
         IsFileOnDisk(f, h->disk) ||
         f->transferable;
}


/*
 * Returns the delay, given the current schedule, until a copy of #f# will be
 * available on #d#; ULONG_MAX if never.
 */
static unsigned long
DelayUntilOnDisk(Disk d,
                 File f) {

  GanttColumn c;
  GanttEvent e;
  unsigned long end;
  FileTransfer ft;
  unsigned i;
  unsigned long now;

  /* Already on #d#? */
  if(IsFileOnDisk(f, d) || IsSpecialFile(f))
    return 0;
  /* On its way to #d#? */
  for(i = 0; i < Size(d->transfersInProgress); i++) {
    ft = d->transfersInProgress[i];
    if(ft->file == f) {
      now = time(NULL);
      end = ft->meta.timeStarted + ft->meta.timeEstimated;
      return now > end ? 0 : (end - now);
    }
  }
  /* Left on #d# by a scheduled task? */
  if(f->outputOf != NULL &&
     (c = GanttChart_GetColumnForDetails(processorChart,f->outputOf)) != NULL &&
     ((Processor)GanttColumn_GetKey(c))->host->disk == d)
    return GanttColumn_GetEventForDetails(c, f->outputOf)->end;
  /* Scheduled for transfer to #d#? */
  if(scheduledInputTransferEvents != NULL) {
    for(i = 0; i < Size(scheduledInputTransferEvents); i++) {
      e = scheduledInputTransferEvents[i];
      ft = (FileTransfer)e->details;
      if(ft->dest == d && ft->file == f)
        return e->end;
    }
  }
  return ULONG_MAX;

}


/*
 * Returns in #bestProcessor# the processor which can complete #t# with the
 * shortest delay.  Returns the delay for the best processor in
 * #bestProcessorDelay# and the difference between the delays on the best and
 * second best processors in #bestProcessorSufferage#.  If #diskEquivalent# is
 * non-zero, all processors that share a disk are considered equivalent for
 * computing the sufferage, i.e., the second best processor will be one that
 * does not share a disk with #bestProcessor#.
 */
static void
FindMinDelays(Task t,
              int diskEquivalent,
              Processor *bestProcessor,
              unsigned long *bestProcessorDelay,
              unsigned long *bestProcessorSufferage) {

  int equivalentToBest;
  unsigned i;
  unsigned long secondBestDelay = ULONG_MAX;
  Processor thisProcessor;
  unsigned long thisProcessorDelay;

  *bestProcessor = NULL;
  *bestProcessorDelay = ULONG_MAX;
  for(i = 0; i < GanttChart_GetColumnCount(processorChart); i++) {
    thisProcessor =
      (Processor)GanttColumn_GetKey(GanttChart_GetColumn(processorChart, i));
    thisProcessorDelay = PlanTask(t, thisProcessor, 0, 0);
    equivalentToBest =
      *bestProcessor == NULL ? 0 :
       (diskEquivalent &&
        (*bestProcessor)->host->disk == thisProcessor->host->disk);
    if(thisProcessorDelay < *bestProcessorDelay) {
      if(!equivalentToBest)
        secondBestDelay = *bestProcessorDelay;
      *bestProcessor = thisProcessor;
      *bestProcessorDelay = thisProcessorDelay;
    }
    else if(!equivalentToBest && thisProcessorDelay < secondBestDelay)
      secondBestDelay = thisProcessorDelay;
  }
  *bestProcessorSufferage = secondBestDelay - *bestProcessorDelay;

}


/*
 * Returns the column from the global linkChart that applies to the link
 * between #d1# and #d2#.  Creates the column if it doesn't already exist.
 */
static GanttColumn
GetLinkColumn(Disk d1,
              Disk d2) {
  GanttColumn c;
  unsigned i;
  Link l;
  for(i = 0; (c = GanttChart_GetColumn(linkChart, i)) != NULL; i++) {
    l = (Link)GanttColumn_GetKey(c);
    if((l->d1 == d1 && l->d2 == d2) || (l->d1 == d2 && l->d2 == d1))
      return c;
  }
  l = (Link)MALLOC(sizeof(struct _Link));
  l->d1 = strcmp(d1->id, d2->id) <= 0 ? d1 : d2;
  l->d2 = l->d1 == d1 ? d2 : d1;
  return GanttChart_AddColumn(linkChart, l);
}


/*
 * Returns the column from the global processorChart that applies to processor
 * #cpu# on host #h#.  Creates the column if it doesn't already exist.
 */
static GanttColumn
GetProcessorColumn(Host h,
                   unsigned cpu) {
  GanttColumn c;
  unsigned i;
  Processor p;
  for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++) {
    p = (Processor)GanttColumn_GetKey(c);
    if(p->host == h && p->cpu == cpu)
      return c;
  }
  p = (Processor)MALLOC(sizeof(struct _Processor));
  p->host = h;
  p->cpu = cpu;
  return GanttChart_AddColumn(processorChart, p);
}


/* Returns 1 iff all the input files for #t# are available on #d#. */
static int
IsInputAvailable(Task t,
                 Disk d) {
  unsigned i;
  for(i = 0; i < Size(t->input); i++) {
    File f = t->input[i];
    if(!IsFileOnDisk(f, d) && !IsSpecialFile(f))
      return 0;
  }
  return 1;
}


/*
 * Returns 1 or 0 depending on whether or not a file transfer is currently
 * active between #d1# and #d2#.
 */
static int
IsLinkBusy(Disk d1,
           Disk d2) {
  FileTransfer ft;
  unsigned i;
  for(i = 0; i < Size(d1->transfersInProgress); i++) {
    ft = d1->transfersInProgress[i];
    if(ft->source == d2 || ft->dest == d2)
      return 1;
  }
  return 0;
}


/* Writes a representation of the Gantt charts to the log file. */
static void
LogGanttCharts(const char *label) {
  GanttColumn c;
  unsigned col;
  char **columnImages;
  unsigned i;
  char image[40];
  char *message;

  columnImages = (char **)
    MALLOC((GanttChart_GetColumnCount(processorChart) +
            GanttChart_GetColumnCount(linkChart) + 1) * sizeof(char *));
  col = 0;
  for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++) {
    Processor p = (Processor)GanttColumn_GetKey(c);
    sprintf(image, ":%ld %ld", (long)p->cpu, GanttColumn_GetEventCount(c));
    columnImages[col++] = StrAppend(NULL, p->host->id, image, NULL);
  }
  for(i = 0; (c = GanttChart_GetColumn(linkChart, i)) != NULL; i++) {
    Link l = (Link)GanttColumn_GetKey(c);
    sprintf(image, " %ld", GanttColumn_GetEventCount(c));
    columnImages[col++] =
      StrAppend(NULL, l->d1->id, "<->", l->d2->id, image, NULL);
  }
  columnImages[col] = NULL;
  message = StrArrayJoin((const char **)columnImages, "\n\t");
  LOG2("%s\n\t%s\n", label, message);
  StrArrayFree(columnImages);
  FREE(message);
}


/*
 * Prints to the log file, prefixed by #label#, a list of indexes from the
 * #size#-long array #values# that have a value of #value#.
 */
static void
LogIndexes(const char *label,
           int value,
           int *values,
           unsigned size) {

  unsigned i;
  char rangeImage[40];
  char *ranges = NULL;
  unsigned startOfRange;

  for(i = 0; i < size; i++) {
    if(values[i] != value)
      continue;
    startOfRange = i;
    for(i++; i < size && values[i] == value; i++)
      ; /* empty */
    if(--i > startOfRange)
      sprintf(rangeImage, "%d-%d,", startOfRange, i);
    else
      sprintf(rangeImage, "%d,", i);
    ranges = StrAppend(ranges, rangeImage, NULL);
  }
  if(ranges == NULL)
    return;
  ranges[strlen(ranges) - 1] = '\0';
  LOG2("%s: %s\n", label, ranges);
  FREE(ranges);

}


/*
 * Returns the delay until a transfer of #file# from #source# to #dest# would
 * complete, assuming it had to wait at least #delayToStart# before starting.
 * Adds entries to the global linkChart if #enterInChart# is true.
 */
static unsigned long
PlanTransfer(Disk source,
             Disk dest,
             File file,
             unsigned long delayToStart,
             int eventType,
             int enterInChart) {

  GanttColumn c = GetLinkColumn(source, dest);
  unsigned long duration;
  GanttEvent e;
  FileTransfer ft;

  ft = FileTransfer_New(file, source, dest);
  if((duration = PredictFileTransfer(ft)) == 0)
    duration = 1; /* 0-length durations cause ordering problems. */
  delayToStart = GanttColumn_EarliestStartTime(c, delayToStart, duration);
  if(enterInChart) {
    ft->meta.timeEstimated = duration;
    e = GanttEvent_New(eventType, delayToStart, delayToStart + duration, ft);
    GanttColumn_AddEvent(c, e);
    if(eventType == INPUT_TRANSFER_EVENT)
      Append(scheduledInputTransferEvents, e);
  }
  else
    FileTransfer_Free(ft);
  return delayToStart + duration;

}


/*
 * Returns the delay until all of #t#'s input files could be transferred to
 * #d#.  Adds entries to the global linkChart if #enterInChart# is true.
 */
static unsigned long
PlanTransferInput(Task t,
                  Disk d,
                  int enterInChart) {

  unsigned long allDone = 0;
  GanttColumn c;
  GanttEvent e;
  unsigned i;
  HashIterator it;
  unsigned long oneDone;
  Disk source;

  if(localDisk == NULL)
    localDisk = FindLocalDisk();
  if (APST_DEBUG_ON == 1) printf("I am after findlocaldisk \n");
  //printf("the localdisk is %s and the task id is %s and the disk id is %s \n", localDisk->id, t->meta.cost, d->id);
  if (APST_DEBUG_ON == 1) printf("the task id is %f and the disk id is %s \n", t->meta.cost, d->id);
  fflush(stdout);
  for(i = 0; i < Size(t->input); i++) {
    File input = t->input[i];
    if((oneDone = DelayUntilOnDisk(d, input)) == ULONG_MAX &&
       input->transferable) {
      if (APST_DEBUG_ON == 1) printf("I am in oneDone \n");
      /* Is a copy available to transfer? */
      source = NULL;
      ForEachMapping(input->copies, it) {
        FileCopy fc = (FileCopy)IteratorValue(it);
        if(fc->disk == localDisk) {
          source = fc->disk;
          break; /* Give local disk preference over all others. */
        }
        if(fc->disk->method == d->method || source == NULL)
          source = fc->disk;
      }
      /* Can we transfer the file after some scheduled task produces it? */
      if(source == NULL &&
         input->outputOf != NULL &&
         (c = GanttChart_GetColumnForDetails(processorChart,input->outputOf)) !=
         NULL) {
        e = GanttColumn_GetEventForDetails(c, input->outputOf);
        if(e->end > allDone)
          allDone = e->end;
        source = ((Processor)GanttColumn_GetKey(c))->host->disk;
      }
      if(source != NULL)
        oneDone = PlanTransfer
          (source, d, input, allDone, INPUT_TRANSFER_EVENT, enterInChart);
    }
    if (APST_DEBUG_ON == 1) printf("I am after oneDone \n");
    
    if(oneDone > allDone)
      allDone = oneDone;
    if(oneDone == ULONG_MAX)
      break; /* No point in continuing. */
  }

  return allDone;

}


/*
 * Returns the delay until all of #t#'s output files could be transferred from
 * #d# to the local disk, assuming that transfers could not begin until
 * #delayToStart#.   Adds entries to the global linkChart if #enterInChart# is
 * true.
 */
static unsigned long
PlanTransferOutput(Task t,
                   Disk d,
                   unsigned long delayToStart,
                   int enterInChart) {
  unsigned long allDone = delayToStart;
  unsigned i;
  unsigned long oneDone;
  if(d->method == ELBASE_CP)
    return delayToStart;
  if(localDisk == NULL)
    localDisk = FindLocalDisk();
  for(i = 0; i < Size(t->output); i++) {
    File output = t->output[i];
    if(!output->download || IsSpecialFile(output))
      continue;
    oneDone = PlanTransfer
      (d, localDisk, output, allDone, OUTPUT_TRANSFER_EVENT, enterInChart);
    if(oneDone > allDone)
      allDone = oneDone;
  }
  return allDone;
}


/*
 * Returns the delay until #t# would complete if executed on #p#.  If
 * #enterInChart# is true, adds entries for the input transfers and task to the
 * globals processorChart and linkChart; if #scheduleOutput# is also true, adds
 * entries for the output transfers as well.
 */
static unsigned long
PlanTask(Task t,
         Processor p,
         int enterInChart,
         int scheduleOutput) {
  GanttColumn c = GanttChart_GetColumnForKey(processorChart, p);
  unsigned long taskDone;
  unsigned long taskDuration;
  unsigned long taskStart;
  if(p->host->memory != 0 && t->meta.memory > p->host->memory)
    return ULONG_MAX;
  if(t->host != NULL && t->host != p->host)
    return ULONG_MAX;
  if(!CompatableExecutable(t->exec, p->host))
    return ULONG_MAX;
  /* Don't schedule any input transfers until we're sure all are available. */
  if (APST_DEBUG_ON == 1) printf("I am before PlanTransferInput \n");
  if (APST_DEBUG_ON == 1) printf("the task cost is %f \n", t->meta.cost);
  if((taskStart = PlanTransferInput(t, p->host->disk, 0)) == ULONG_MAX)
    return ULONG_MAX;
  if (APST_DEBUG_ON == 1) printf("I am after PlanTransferInput \n");
  if(enterInChart)
    PlanTransferInput(t, p->host->disk, 1);
  if (APST_DEBUG_ON == 1) printf("I am after PlanTransferInput 1\n");
  if((taskDuration = PredictTaskOnHost(t, p->host)) == 0)
    taskDuration = 1; /* 0-length durations cause ordering problems. */
  taskStart = GanttColumn_EarliestStartTime(c, taskStart, taskDuration);
  taskDone = taskStart + taskDuration;
  if(enterInChart) {
    t->meta.timeEstimated = taskDuration;
    GanttColumn_AddEvent(c, GanttEvent_New(TASK_EVENT, taskStart, taskDone, t));
    if(scheduleOutput)
      PlanTransferOutput(t, p->host->disk, taskDone, 1);
  }
  return taskDone;
}


/*
 * Returns in #t# and #p# the best task/processor pairing under #heuristic# for
 * tasks selected from the #len#-long array #choices#.
 */
static void
SelectBestPair(heuristic_t heuristic,
               Task *choices,
               unsigned len,
               Task *t,
               Processor *p) {

  GanttColumn c;
  unsigned i;
  Processor proc;

  *p = NULL;
  *t = NULL;
  if(len == 0)
    return;

  if(heuristic == HEURISTIC_ROUNDROBIN || heuristic == HEURISTIC_WORK_QUEUE) {
    unsigned col = len % GanttChart_GetColumnCount(processorChart);
    unsigned firstCol = col;
    do {
      c = GanttChart_GetColumn(processorChart, col);
      if(heuristic == HEURISTIC_ROUNDROBIN || GanttColumn_IsEmpty(c)) {
        proc = (Processor)GanttColumn_GetKey(c);
        for(i = 0; i < len && PlanTask(choices[i],proc,0,0) == ULONG_MAX; i++)
          ; /* empty */
        if(i < len) {
          *p = proc;
          *t = choices[i];
        }
      }
      if(++col == GanttChart_GetColumnCount(processorChart))
        col = 0;
    } while(*p == NULL && col != firstCol);
  }
  else {
    unsigned long bestScore = 0;
    int byDisk = heuristic == HEURISTIC_XSUFFERAGE;
    Processor minTaskProcessor;
    unsigned long minTaskProcessorTime;
    unsigned long sufferage;
    unsigned long taskScore;
    if(len > MAX_TASKS)
      len = MAX_TASKS;
    for(i = 0; i < len; i++) {
      FindMinDelays(choices[i], byDisk, &minTaskProcessor,
                    &minTaskProcessorTime, &sufferage);
      taskScore =
        heuristic == HEURISTIC_MINMIN ? (ULONG_MAX - minTaskProcessorTime) :
        heuristic == HEURISTIC_MAXMIN ? minTaskProcessorTime : sufferage;
      if(taskScore >= bestScore) {
        *p = minTaskProcessor;
        *t = choices[i];
        bestScore = taskScore;
      }
    }
  }

}


/* Fills the linkChart and processorChart globals with a schedule. */
static void
FillGanttCharts(void) {

  GanttColumn c;
  unsigned cpu;
  GanttEvent e;
  unsigned long end;
  unsigned highestPriority;
  SequenceNumber highestSeq;
  unsigned i;
  HashIterator it;
  unsigned leftToSchedule;
  TaskList newTasks;
  unsigned long now = time(NULL);
  Processor p;
  TaskList runningTasks;
  TaskSchedulability *schedulability;
  Task *unscheduled;
  Task t;
  unsigned long tEnd;

  Host *activeHosts; 
  Host h; 
  int taskNum, hostNum, procNum;
  HostParameters procParas[MAX_WS_NUM]; //follow mtsc.c definition
  double utilRates[MAX_WS_NUM], arriRates[MAX_WS_NUM], servStds[MAX_WS_NUM], speedRates[MAX_WS_NUM];
  double taskWorkloads[MAX_GJOB_NUM];
  int maps[MAX_GJOB_NUM];

  //unsigned i;
  Processor proc;
  
  /* Add placeholders for transfers and tasks already under way. */
  ForEachMapping(global.transfersInProgress, it) {
    FileTransfer ft = (FileTransfer)IteratorValue(it);
    if((end = ft->meta.timeStarted + ft->meta.timeEstimated) > now)
      GanttColumn_AddEvent(GetLinkColumn(ft->source, ft->dest),
                           GanttEvent_New(OUTPUT_TRANSFER_EVENT, 0, end - now,
                                          NULL));
  }
  runningTasks = GetValue(global.taskGroups, "running", NULL);
  ForEachMapping(runningTasks, it) {
    t = (Task)IteratorValue(it);
    if((end = t->meta.timeStarted + t->meta.timeEstimated) > now) {
      for(cpu = 0; cpu < t->host->cpus; cpu++) {
        c = GetProcessorColumn(t->host, cpu);
        if(!GanttColumn_IsEmpty(c))
          continue;
        GanttColumn_AddEvent(c, GanttEvent_New(TASK_EVENT, 0, end - now, NULL));
        break;
      }
    }
    PlanTransferOutput
      (t, t->host->disk,
       !transferOutputImmediately ? 10000 : end > now ? end - now : 0, 1);
  }

  /* Add tasks (and associated file transfers) that are still to be done. */
  highestPriority = 0;
  highestSeq = 0;
  newTasks = GetValue(global.taskGroups, "new", NULL);
  ForEachMapping(newTasks, it) {
    t = (Task)IteratorValue(it);
    if(t->priority > highestPriority)
      highestPriority = t->priority;
    if(t->seq > highestSeq)
      highestSeq = t->seq;
  }
  schedulability = MALLOC((highestSeq + 1) * sizeof(TaskSchedulability));
  memset(schedulability, NOT_NEW, (highestSeq+1) * sizeof(TaskSchedulability));
  unscheduled = VectorNew(Task);
  SetCapacity(unscheduled, GetMappingCount(newTasks));
  ForEachMapping(newTasks, it) {
    t = (Task)IteratorValue(it);
    if((schedulability[t->seq] = CheckTaskSchedulability(t, highestPriority)) ==
        SCHEDULABLE)
      Append(unscheduled, t);
  }

  activeHosts = VectorNew(Host);
  SetCapacity(activeHosts, 100);
  ForEachMapping(global.hosts, it) {
      h = (Host)IteratorValue(it);
      if(h->meta.timeDisabled != 0)
        continue;
      Append(activeHosts, h);
  }

  LogIndexes("Tasks unready; host disabled",
             HOST_DISABLED, (int *)schedulability, highestSeq);
  LogIndexes("Tasks unready; input unavailable",
             INPUT_UNAVAILABLE, (int *)schedulability, highestSeq);
  LogIndexes("Tasks unready; input inaccessible",
             INPUT_INACCESSIBLE, (int *)schedulability, highestSeq);
  LogIndexes("Tasks unready; insufficient memory",
             INSUFFICIENT_MEMORY, (int *)schedulability, highestSeq);
  LogIndexes("Tasks unready; priority too low",
             LOW_PRIORITY, (int *)schedulability, highestSeq);
  LogIndexes("Tasks ready", SCHEDULABLE, (int *)schedulability, highestSeq);
  FREE(schedulability);

  if (global.heuristic != HEURISTIC_GHS) {
  if (APST_DEBUG_ON == 1) printf("I am using APST to schedule tasks ****************** \n"); 
  /* Arbitrarily reschedule after doing an avg of two tasks on each proc. */
  leftToSchedule = GanttChart_GetColumnCount(processorChart) * 2;
  if(leftToSchedule > Size(unscheduled))
    leftToSchedule = Size(unscheduled);
  for( ; leftToSchedule > 0; leftToSchedule--) {
    SelectBestPair(global.heuristic, unscheduled, Size(unscheduled), &t, &p);
    if(t == NULL || p == NULL)
      break;
    tEnd = PlanTask(t, p, 1, transferOutputImmediately);
    for(i = 0; unscheduled[i] != t; i++)
      ; // empty 
    Remove(unscheduled, i);
  }
  VectorFree(unscheduled);
  VectorFree(activeHosts);

  } else {

  if (APST_DEBUG_ON == 1) printf("I am using GHS to schedule tasks ****************** \n"); 
  scheduleFrequency = 7200;  
  taskNum = Size(unscheduled);
  hostNum = Size(activeHosts);
  if (DEBUG_ON == 1) printf("the task num is %d, and the host num is %d \n", taskNum, hostNum);
  if (taskNum > 0) {
  for (i = 0; i < taskNum; i++) {
	taskWorkloads[i] = unscheduled[i]->meta.cost;
	//printf("the workload of task %d is %f \n", i, unscheduled[i]->meta.cost);
  }
  if (DEBUG_ON == 1) printf("i am at before GetParasValue\n");
  procNum = GetParasValue(utilRates, arriRates, servStds, speedRates, hostNum, activeHosts);
  if (DEBUG_ON == 1) printf("i am after GetParasValue\n");
  for (i = 0; i< procNum; i++) {
    procParas[i].utilRate = utilRates[i];
    procParas[i].arriRate = arriRates[i];
    procParas[i].servStd = servStds[i]; 
    procParas[i].speedRate = speedRates[i]; 
  }

  //collect tasks' workloads information and machine parameters information
  if (DEBUG_ON == 1) printf("i am before GhsMetataskSched\n");
  GhsMetataskSched(taskWorkloads, taskNum, procParas, procNum, maps);
  if (DEBUG_ON == 1) printf("i am after GhsMetataskSched\n");
  for(i = 0; i < taskNum; i++) {
    if (APST_DEBUG_ON == 1) printf("the task %d with a cost of %f is allocated to machine %s \n", i, unscheduled[i]->meta.cost, activeHosts[maps[i]]->id);
  }
  for(i = 0; i < taskNum; i++) {
    c = GanttChart_GetColumn(processorChart, maps[i]);
    proc = (Processor)GanttColumn_GetKey(c);
    //tEnd = PlanTask(unscheduled[i], activeHosts[maps[i]], 1, transferOutputImmediately);
    tEnd = PlanTask(unscheduled[i], proc, 1, transferOutputImmediately);
  }

  } // end of if 
  VectorFree(unscheduled);
  VectorFree(activeHosts);
  if (DEBUG_ON == 1) printf("i am after PlanTask\n");

  }

  /* Remove any placeholders. */
  for(i = 0; (c = GanttChart_GetColumn(linkChart, i)) != NULL; i++)
    while(!GanttColumn_IsEmpty(c) &&
          (e = GanttColumn_GetFirstEvent(c))->details == NULL)
      GanttEvent_Free(GanttColumn_RemoveEvent(c, e));
  for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++)
    while(!GanttColumn_IsEmpty(c) &&
          (e = GanttColumn_GetFirstEvent(c))->details == NULL)
      GanttEvent_Free(GanttColumn_RemoveEvent(c, e));

  /* Add output transfers for all scheduled tasks. */
  if(!transferOutputImmediately) {
    for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++) {
      p = (Processor)GanttColumn_GetKey(c);
      GanttColumn_ForEachEvent(c, e)
        PlanTransferOutput((Task)e->details, p->host->disk, e->end, 1);
    }
  }

}


int
sched_file_transfer_done(FileTransfer ft,
                         int succeeded) {
  return 1;
}


int
sched_host_available(Host h) {
  sched_reschedule();
  return 1;
}


int
sched_host_failed(Host h) {
  sched_reschedule();
  return 1;
}


int
sched_init(void) {
  return 1;
}


int
sched_new_task(Task t) {
  sched_reschedule();
  return 1;
}


int
sched_reschedule(void) {
  expires = 0;
  return 1;
}


int
sched_run(int execute) {

  Disk bulkSource;
  FileTransferList bulkTransfers;
  GanttColumn c;
  unsigned cpu;
  GanttEvent e;
  File f;
  FileCopy fc;
  FileTransfer ft;
  Host h;
  HashTable hostStarts;
  unsigned i;
  HashIterator it;
  Link l;
  GanttEvent nextE;
  FileTransfer oldFt;
  Task t;
  int thisHostStarts;

  /* If our schedule is expired, make a new one if we have tasks and hosts. */
  if(time(NULL) >= expires) {

    if(processorChart != NULL) {
      for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++)
        FREE(GanttColumn_GetKey(c));
      GanttChart_Free(processorChart);
    }
    if(linkChart != NULL) {
      for(i = 0; (c = GanttChart_GetColumn(linkChart, i)) != NULL; i++) {
        while(!GanttColumn_IsEmpty(c)) {
          e = GanttColumn_RemoveEvent(c, GanttColumn_GetFirstEvent(c));
          FileTransfer_Free((FileTransfer)e->details);
          GanttEvent_Free(e);
        }
        FREE(GanttColumn_GetKey(c));
      }
      GanttChart_Free(linkChart);
    }
    processorChart = GanttChart_New();
    linkChart = GanttChart_New();
    /* Add transfers for any stranded output files. */
    ForEachMapping(global.files, it) {
      f = (File)IteratorValue(it);
      if(f->outputOf == NULL || !f->download || GetMappingCount(f->copies) == 0)
        continue;
      if(localDisk == NULL)
        localDisk = FindLocalDisk();
      if(DelayUntilOnDisk(localDisk, f) != ULONG_MAX)
        continue;
      fc = (FileCopy)IteratorValue(IteratorFirst(f->copies));
      PlanTransfer
        (fc->disk, localDisk, f, transferOutputImmediately ? 0 : 10000,
         OUTPUT_TRANSFER_EVENT, 1);
    }

    /* Find out how many active hosts we have. */
    ForEachMapping(global.hosts, it) {
      h = (Host)IteratorValue(it);
      if(h->meta.timeDisabled != 0)
        continue;
      for(cpu = 0; cpu < h->cpus; cpu++)
        GetProcessorColumn(h, cpu);
    }
    if(GanttChart_GetColumnCount(processorChart) > 0) {
      unsigned long started = time(NULL);
      scheduledInputTransferEvents = VectorNew(GanttEvent);
      SetCapacity(scheduledInputTransferEvents, 100);
      SetIncrement(scheduledInputTransferEvents, 100);
      FillGanttCharts();
      VectorFree(scheduledInputTransferEvents);
      scheduledInputTransferEvents = NULL;
      LOG1("Finished scheduling in %ld seconds\n", time(NULL) - started);
    }

    if(!GanttChart_IsEmpty(processorChart) || !GanttChart_IsEmpty(linkChart))
      LogGanttCharts("Reschedule");
    expires = time(NULL) + scheduleFrequency;

  }

  if(!execute)
    return 1;

  /* Start the next file transfer on each link ... */
  for(i = 0; (c = GanttChart_GetColumn(linkChart, i)) != NULL; i++) {
    l = (Link)GanttColumn_GetKey(c);
    if(GanttColumn_IsEmpty(c) || IsLinkBusy(l->d1, l->d2))
      continue;
    bulkSource = NULL;
    bulkTransfers = FileTransferList_New(LINKED_LIST);
    for(e = GanttColumn_GetFirstEvent(c); e != NULL; e = nextE) {
      nextE = GanttColumn_GetNextEvent(c, e);
      ft = (FileTransfer)e->details;
      if((bulkSource != NULL && ft->source != bulkSource) ||
         !IsFileOnDisk(ft->file, ft->source))
        continue;
      if((oldFt = FileTransferList_Get(global.transfers, ft->id)) == NULL)
        FileTransferList_Add(global.transfers, ft);
      else {
        FileTransfer_Free(ft);
        FileTransferReset(oldFt);
        ft = oldFt;
      }
      FileTransferList_Add(bulkTransfers, ft);
      bulkSource = ft->source;
      GanttEvent_Free(GanttColumn_RemoveEvent(c, e));
      if(!global.bulkCopy)
        break;
    }
    storage_do_bulk_transfer(bulkTransfers);
    FileTransferList_Free(bulkTransfers, 0);
  }

  /* ... and the next task on each processor. */
  hostStarts = HashTableNew(SMALL_HASH, NULL, NULL, NULL);
  for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++) {
    h = ((Processor)GanttColumn_GetKey(c))->host;
    thisHostStarts = (int)GetValue(hostStarts, h, NULL);
    if(GanttColumn_IsEmpty(c) || Size(h->tasksInProgress) >= h->cpus ||
       thisHostStarts >= MAX_CONCURRENT_HOST_STARTS)
      continue;
    GanttColumn_ForEachEvent(c, e) {
      t = (Task)e->details;
      if(!IsInputAvailable(t, h->disk))
        continue;
      if(e != GanttColumn_GetFirstEvent(c))
        WARN1("Bypassing unready task %ld\n",
              ((Task)GanttColumn_GetFirstEvent(c)->details)->seq);
      t->host = h;
      compute_do_task(t);
      GanttEvent_Free(GanttColumn_RemoveEvent(c, e));
      break;
    }
    PutMapping(hostStarts, h, (KeyType)(thisHostStarts + 1));
  }
  HashTableFree(hostStarts);

  return 1;

}


int
sched_shutdown(void) {
  GanttColumn c;
  GanttEvent e;
  unsigned i;
  if(linkChart != NULL) {
    for(i = 0; (c = GanttChart_GetColumn(linkChart, i)) != NULL; i++) {
      GanttColumn_ForEachEvent(c, e)
        FileTransfer_Free((FileTransfer)e->details);
      FREE(GanttColumn_GetKey(c));
    }
    GanttChart_Free(linkChart);
  }
  if(processorChart != NULL) {
    for(i = 0; (c = GanttChart_GetColumn(processorChart, i)) != NULL; i++)
      FREE(GanttColumn_GetKey(c));
    GanttChart_Free(processorChart);
  }
  linkChart = processorChart = NULL;
  sched_reschedule();
  localDisk = NULL;
  return 1;
}


int
sched_task_canceled(Task t) {
  sched_reschedule(); /* In case we have dependencies */
  return 1;
}


int
sched_task_done(Task t,
                int succeeded) {
  if(processorChart != NULL) {
    if(!succeeded)
      sched_reschedule(); /* In case we have dependencies */
    else {
      /* Reschedule if this was the final task scheduled for this processor. */
      unsigned i;
      for(i = 0;
          i < t->host->cpus &&
          !GanttColumn_IsEmpty(GetProcessorColumn(t->host, i));
          i++)
        ; /* empty */
      if(i < t->host->cpus)
        sched_reschedule();
    }
  }
  return 1;
}
